@@ -1,31 +1,16 @@
 import radcad.core as core
 import radcad.wrappers as wrappers
+from radcad.backends import Backend
 from radcad.utils import flatten, extract_exceptions
 
 import multiprocessing
-
-# Pathos re-writes the core code in Python rather than C, for ease of maintenance at cost of performance
-from pathos.multiprocessing import ProcessPool as PathosPool
 import dill
-
-import ray
-
-from enum import Enum
 import copy
 
 
+# Get machine CPU count
 cpu_count = multiprocessing.cpu_count() - 1 or 1
 
-
-class Backend(Enum):
-    DEFAULT = 0
-    MULTIPROCESSING = 1
-    RAY = 2
-    RAY_REMOTE = 3
-    PATHOS = 4
-    SINGLE_PROCESS = 5
-
-
 class Engine:
     def __init__(self, **kwargs):
         self.experiment = None
@@ -34,6 +19,7 @@ class Engine:
         self.raise_exceptions = kwargs.pop("raise_exceptions", True)
         self.deepcopy = kwargs.pop("deepcopy", True)
         self.drop_substeps = kwargs.pop("drop_substeps", False)
+        self.run_generator = iter(())
 
         if kwargs:
             raise Exception(f"Invalid Engine option in {kwargs}")
@@ -64,88 +50,29 @@ class Engine:
 
         self.experiment._before_experiment(experiment=self.experiment)
 
-        run_generator = self._run_stream(configs)
+        self.run_generator = self._run_stream(configs)
 
+        # Select backend executor
         if self.backend in [Backend.RAY, Backend.RAY_REMOTE]:
             if self.backend == Backend.RAY_REMOTE:
-                print(
-                    "Using Ray remote backend, please ensure you've initialized Ray using ray.init(address=***, ...)"
-                )
+                from radcad.extensions.backends.ray import ExecutorRayRemote as Executor
             else:
-                ray.init(num_cpus=self.processes, ignore_reinit_error=True)
-
-            futures = [
-                Engine._proxy_single_run_ray.remote((config, self.raise_exceptions))
-                for config in run_generator
-            ]
-            result = ray.get(futures)
+                from radcad.extensions.backends.ray import ExecutorRay as Executor
         elif self.backend in [Backend.PATHOS, Backend.DEFAULT]:
-            with PathosPool(self.processes) as pool:
-                result = pool.map(
-                    Engine._proxy_single_run,
-                    [
-                        (config, self.raise_exceptions)
-                        for config in run_generator
-                    ],
-                )
-                pool.close()
-                pool.join()
-                pool.clear()
+            from radcad.backends.pathos import ExecutorPathos as Executor
         elif self.backend in [Backend.MULTIPROCESSING]:
-            with multiprocessing.get_context("spawn").Pool(
-                processes=self.processes
-            ) as pool:
-                result = pool.map(
-                    Engine._proxy_single_run,
-                    [
-                        (config, self.raise_exceptions)
-                        for config in run_generator
-                    ],
-                )
-                pool.close()
-                pool.join()
+            from radcad.backends.multiprocessing import ExecutorMultiprocessing as Executor
         elif self.backend in [Backend.SINGLE_PROCESS]:
-            result = [
-                Engine._proxy_single_run((config, self.raise_exceptions))
-                for config in run_generator
-            ]
+            from radcad.backends.single_process import ExecutorSingleProcess as Executor
         else:
             raise Exception(f"Execution backend must be one of {Backend._member_names_}, not {self.backend}")
         
+        result = Executor(self).execute_runs()
+        
         self.experiment.results, self.experiment.exceptions = extract_exceptions(result)
         self.experiment._after_experiment(experiment=self.experiment)
         return self.experiment.results
 
-    @ray.remote
-    def _proxy_single_run_ray(args):
-        return Engine._proxy_single_run(args)
-    
-    def _proxy_single_run(args):
-        return Engine._single_run(args)
-
-    def _single_run(args):
-        run_args, raise_exceptions = args
-        try:
-            results, exception, traceback = core.single_run(*tuple(run_args))
-            if raise_exceptions and exception:
-                raise exception
-            else:
-                return results, {
-                        'exception': exception,
-                        'traceback': traceback,
-                        'simulation': run_args.simulation,
-                        'run': run_args.run,
-                        'subset': run_args.subset,
-                        'timesteps': run_args.timesteps,
-                        'parameters': run_args.parameters,
-                        'initial_state': run_args.initial_state,
-                    }
-        except Exception as e:
-            if raise_exceptions:
-                raise e
-            else:
-                return [], e
-
     def _get_simulation_from_config(config):
         states, state_update_blocks, params, timesteps, runs = config
         model = wrappers.Model(
@@ -204,7 +131,7 @@ class Engine:
                             self.drop_substeps,
                         )
                         self.experiment._after_subset(context=context)
-                    self.experiment._before_run(context=context)
+                    self.experiment._after_run(context=context)
                 else:
                     context = wrappers.Context(
                         simulation_index,
@@ -215,6 +142,7 @@ class Engine:
                         params
                     )
                     self.experiment._before_run(context=context)
+                    self.experiment._before_subset(context=context)
                     yield wrappers.RunArgs(
                         simulation_index,
                         timesteps,
@@ -226,6 +154,7 @@ class Engine:
                         self.deepcopy,
                         self.drop_substeps,
                     )
+                    self.experiment._after_subset(context=context)
                     self.experiment._after_run(context=context)
 
             self.experiment._after_simulation(
